#include "AVSReader.h"
#pragma comment(lib, "vfw32.lib")
#pragma comment(lib, "Ole32.lib")
/*
#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif
*/
#define     GlobalAllocPtr(flags, cb)        \
                (GlobalLock(GlobalAlloc((flags), (cb))))

CAVSReader::CAVSReader()
{
	CoInitialize(NULL);
	pAVIFileIn		= NULL;
	pVideoStream	= NULL;
	pAudioStream	= NULL;
	m_pAudioFrame	= NULL;
	m_pVideoFrame	= NULL;
	m_bHasAudio		= false;
	m_bHasVideo		= false;

	// Init the AVI
	AVIFileInit();
}

CAVSReader::~CAVSReader()
{
	CloseAVSFile();
	AVIFileExit();
	CoUninitialize();
}

void CAVSReader::CloseAVSFile()
{
	if ( m_pAudioFrame )
	{
		free(m_pAudioFrame);
		m_pAudioFrame = NULL;
	}

	if ( m_pVideoFrame )
	{
		free(m_pVideoFrame);
		m_pVideoFrame = NULL;
	}

	if (pVideoStream)
	{
		AVIStreamClose(pVideoStream);
		pVideoStream = NULL;
	}

	if (pAudioStream)
	{
		AVIStreamClose(pAudioStream);
		pAudioStream = NULL;
	}

	if (pAVIFileIn)
	{
		AVIFileClose(pAVIFileIn);
		pAVIFileIn = NULL;
	}
}

int CAVSReader::GetNumFrames(int nCaps)
{
	int nRet = 0;

	switch ( nCaps )
	{
	case 0: // Not Requested
	case 1:	// Only Video 
		nRet = m_nTotalVideoFrames;
		break;
	case 2:	// Only Audio 
		nRet = m_nTotalAudioFrames;
		break;
	case 3: // Both
		nRet = m_nTotalVideoFrames;
		break;
	}

	return nRet;
}

int CAVSReader::GetWidth()
{
	return biyuv.biWidth;
}

int CAVSReader::GetHeight()
{
	return biyuv.biHeight;
}

DWORD CAVSReader::GetScale()
{
	return dwScale;
}


DWORD CAVSReader::GetRate()
{
	return dwRate;
}

extern "C" const GUID CLSID_CAVIFileSynth   // {E6D6B700-124D-11D4-86F3-DB80AFD98778}
  = {0xe6d6b700, 0x124d, 0x11d4, {0x86, 0xf3, 0xdb, 0x80, 0xaf, 0xd9, 0x87, 0x78}};

extern "C" const GUID IID_IAvisynthClipInfo   // {E6D6B708-124D-11D4-86F3-DB80AFD98778}
  = {0xe6d6b708, 0x124d, 0x11d4, {0x86, 0xf3, 0xdb, 0x80, 0xaf, 0xd9, 0x87, 0x78}};


struct IAvisynthClipInfo : IUnknown {
  virtual int __stdcall GetError(const char** ppszMessage) = 0;
  virtual bool __stdcall GetParity(int n) = 0;
  virtual bool __stdcall IsFieldBased() = 0;
};

IAvisynthClipInfo *pAvisynthClipInfo;

bool CAVSReader::OpenAVSFile(const char *szAVSFile)
{
	int Err, oErr;
	
	nFrameCount = 0;
	nAudioFrameCount = 0;

	m_bHasAudio = false;
	m_bHasVideo = false;

	// AVS Reader
	oErr = AVIFileOpen(&pAVIFileIn, szAVSFile, OF_READ | OF_SHARE_DENY_NONE, NULL);

	if ( !oErr ) 
	{
		Err = AVIFileGetStream( pAVIFileIn, &pVideoStream, streamtypeVIDEO, 0);

		if (FAILED(pAVIFileIn->QueryInterface(IID_IAvisynthClipInfo, (void **)&pAvisynthClipInfo)))
			pAvisynthClipInfo = NULL;
		else 
		{
			const char *s;

			if (pAvisynthClipInfo->GetError(&s)) 
			{
				printf("Avisynth open failure:\n%s", s);
				pAvisynthClipInfo->Release();
				pAVIFileIn->Release();
				pAVIFileIn = NULL;

				return false;
			}

			pAvisynthClipInfo->Release();
		}

		if ( !Err ) 
		{
			Err = AVIStreamInfo( pVideoStream, &videoStreamInfo, sizeof( videoStreamInfo ) );

			if ( !Err )
			{
				LONG			cbBMI;
				BITMAPINFO		*pBMI;
		
				AVIStreamFormatSize( pVideoStream, 0, &cbBMI );
				pBMI = (BITMAPINFO *)new BYTE[ cbBMI ];
				AVIStreamReadFormat( pVideoStream, 0, pBMI, &cbBMI );
				memcpy(&m_BMI, &pBMI->bmiHeader, sizeof(BITMAPINFOHEADER));
				delete [] pBMI;

				m_bHasVideo = true;
			}
		}
	}

	// Audio Stream Stuff (don't error if it doesn't work)
	if ( !oErr )
	{
		if ( pAudioStream )
		{
			free(pAudioStream);
			pAudioStream = NULL;
		}

		Err = AVIFileGetStream( pAVIFileIn, &pAudioStream, streamtypeAUDIO, 0);

		if ( !Err )
		{
			m_pAudioFrame = (BYTE*)malloc(AUDIO_BUF_SIZE+100);

			Err = AVIStreamInfo( pAudioStream, &audioStreamInfo, sizeof( audioStreamInfo ) );

			LONG wavFmtSize = 0;

			AVIStreamFormatSize(pAudioStream, 0, &wavFmtSize );
			AVIStreamReadFormat(pAudioStream, 0, &wavFmt, &wavFmtSize);

			m_bHasAudio = true;
		}
	}

	if ( m_bHasAudio == false && m_bHasVideo == false )
		return false;

	if ( m_bHasVideo )
	{
		// Setup YUV Bitmap
		ZeroMemory(&biyuv, sizeof(BITMAPINFOHEADER));
		biyuv = m_BMI; 

		if ( m_pVideoFrame )
		{
			free(m_pVideoFrame);
			m_pVideoFrame = NULL;
		}

		m_pVideoFrame = (BYTE*) malloc(biyuv.biSizeImage*2);

		dwScale = videoStreamInfo.dwScale;
		dwRate = videoStreamInfo.dwRate;
	}
	else
	{
		dwScale = audioStreamInfo.dwScale;
		dwRate = audioStreamInfo.dwRate;
	}

	if ( m_bHasVideo )
		m_nTotalVideoFrames = videoStreamInfo.dwLength;	
	
	if ( m_bHasAudio )
		m_nTotalAudioFrames = audioStreamInfo.dwLength;

	return true;
}

bool CAVSReader::HasAudio()
{
	return m_bHasAudio;
}

bool CAVSReader::HasVideo()
{
	return m_bHasVideo;
}

int CAVSReader::GetChannels()
{
	return wavFmt.wf.nChannels;
}

int CAVSReader::GetSampleRate()
{
	return wavFmt.wf.nSamplesPerSec;
}

int CAVSReader::GetBitsPerSec()
{
	return (int)wavFmt.wBitsPerSample;
}

PCMWAVEFORMAT* CAVSReader::GetWF()
{
	return &wavFmt;
}

void CAVSReader::GetAudioSamplesToRead(int nSampleSize)
{
	//int nAudioSampleSize = /*wavFmt.wf.nChannels **/ (wavFmt.wBitsPerSample/8);

	m_nAudioSamplesToRead = nSampleSize;//(nSampleSize*/*wavFmt.wf.nChannels*/2)/nAudioSampleSize;

//	// If only audio adjust number of frames
//	if ( m_bHasAudio && !m_bHasVideo )
		m_nTotalAudioFrames /= m_nAudioSamplesToRead;
}

BYTE * CAVSReader::GetFrame(LONG *yuy2c_size, LONG *samples)
{
	if ( AVIStreamRead(pVideoStream, nFrameCount++, 1, m_pVideoFrame, biyuv.biSizeImage, yuy2c_size, samples) )
		printf("Read Failed \n");
	if ( nFrameCount >= m_nTotalVideoFrames )
		return NULL;
	else
		return m_pVideoFrame;
}

BYTE * CAVSReader::GetFrame(int nFrame, LONG *yuy2c_size, LONG *samples)
{
	if ( AVIStreamRead(pVideoStream, nFrame, 1, m_pVideoFrame, biyuv.biSizeImage, yuy2c_size, samples) )
		printf("Read Failed \n");
	return m_pVideoFrame;
}

LONG CAVSReader::GetAudioFrame()
{
	LONG read;
	LONG bytes;
	INT e = AVIStreamRead(pAudioStream, nAudioFrameCount, m_nAudioSamplesToRead, m_pAudioFrame, AUDIO_BUF_SIZE, &bytes, &read);
	nAudioFrameCount+=m_nAudioSamplesToRead;
	return bytes;
}